import java.io.FileWriter;
import java.io.IOException;
import java.util.*;

public class MM1Queue {

    private double lambda;
    private double mu;
    private Queue<Event> eventQueue;
    private List<Double> arrivalTimesList;
    private List<Double> departureTimesList;
    private final Random random = new Random();

    public MM1Queue(double lambda, double mu){
        this.lambda = lambda;
        this.mu = mu;
    }

    public MM1Queue(){
        this(0d, 0d);
    }

    /**
     * @param D
     * Cette fonction permet de simulé les instance d'arrivée et de départ
     * sur une période D donnée.
     */
    public void simulate(double D) {
        this.eventQueue = new PriorityQueue<>();

       this.arrivalTimesList = new ArrayList<>();
        this.departureTimesList = new ArrayList<>();

        double currentTime = 0;
        double totalArrivals = 0;
        double totalDepartures = 0;

        while (currentTime < D) {
            // Générer les évenéments d'arrivé
            double arrivalTime = currentTime + this.getNextArrivalTime();
            if (arrivalTime <= D) {
                this.eventQueue.add(new Event(arrivalTime, EventType.ARRIVAL));
                totalArrivals++;
            }

            // Générer les départs
            double departureTime = currentTime + this.getNextDepartureTime();
            if (departureTime <= D) {
                eventQueue.add(new Event(departureTime, EventType.DEPARTURE));
                totalDepartures++;
            }

            // Traiter le prochain événement
            Event event = this.eventQueue.poll();
            currentTime = arrivalTime;

            if (event.getType() == EventType.ARRIVAL) {
                arrivalTimesList.add(currentTime);
                //System.out.println("Temps d'arrivé " + currentTime);
            } else if (event.getType() == EventType.DEPARTURE) {
                //System.out.println("Temps de départ " + currentTime);
                departureTimesList.add(currentTime);
            }
        }

        System.out.println("Nombre total d’arrivées : " + totalArrivals);
        System.out.println("nombre total de départs : " + totalDepartures);
    }

    /**
     * @param D
     * @param filename
     * @throws IOException
     * cette fonction permet de sauvégarder les données de simulation dans un fichier
     */
    public void simulateAndSave(double D, String filename) throws IOException {
        this.simulate(D);
        try (FileWriter writer = new FileWriter(filename)) {
            //ecrire les information pour chaque fichier
            writer.write(String.format("# lambda = %.3f, mu = %.3f%n", this.lambda, this.mu));
            writer.write("# arrival : departure \n");
            // Boucle pour écrire les données dans le fichier
            for (int i = 0; i < Math.min(this.arrivalTimesList.size(), this.departureTimesList.size()); i++) {
                writer.write(String.format("%d %.3f %.3f%n", i, this.arrivalTimesList.get(i),
                        this.departureTimesList.get(i)));
            }
        }
    }

    private  double getNextArrivalTime() {
        return - (1 / this.lambda) * Math.log(1 - random.nextDouble());
    }

    private double getNextDepartureTime() {
        return - (1 / this.mu) * Math.log(1 - random.nextDouble());
    }

    public void setLambda(double lambda) {
        this.lambda = lambda;
    }

    public void setMu(double mu) {
        this.mu = mu;
    }
}
